#include "FastGraphCluster.h"
#include "FibonacciHeap.h"
#include "DegreeArray.h"
#include <cstdlib>

#define UNEXPLORED -1
#define DEBUG_MODE
#define CLST_TOJOIN 0.3
#define getmax(a,b) ((a)>(b)?(a):(b))
#define getmin(a,b) ((a)<(b)?(a):(b))

extern vector<string> vertexName;
extern float cur_pos_start;
extern float cur_pos;
extern string clstmode;
extern float WINDOW_SIZE;

void update0(map<int, size_t> &clst, int k, size_t p)
{
  map<int, size_t>::iterator i;
  if ( (i=clst.find(k)) == clst.end()){
    clst[k] = p;
  } else
    i->second = getmax(i->second, p);
}

void update1(map<int, size_t> &clst, int k, size_t p)
{
  map<int, size_t>::iterator i;
  if ( (i=clst.find(k)) == clst.end()){
    clst[k] = p;
  } else {
    i->second = getmin(i->second, p);
  }
}

FastGraphCluster::FastGraphCluster(float density,int lowersize,float lowerincrease, int m_nVertex):m_nLowerSize(lowersize),m_dLowDen(density),m_dLowerIncrease(lowerincrease), m_nVertex(m_nVertex), clst_topindex(0)
{
	int i;
	m_neighbor = new map<int, EdgeInfo* >[m_nVertex]; 
	neighborWeightCnt = new float[m_nVertex];
	m_pHeapNode = new FiboNode*[m_nVertex];
	m_pHeapNode2 = new FiboNode*[m_nVertex];
	for (i=0;i<m_nVertex;i++)
	  {
	    m_pHeapNode[i] = new FiboNode(HeapNode(UNEXPLORED,i));
	    m_pHeapNode2[i] = m_pHeapNode[i];
	    clstID[i] = -1;
	  }
}

FastGraphCluster::~FastGraphCluster(void)
{
  for (int i=0; i < m_nVertex; i++)
    {
      delete m_pHeapNode[i];
    }
    delete[] m_pHeapNode; // FiboNode **
    delete[] m_pHeapNode2;
    delete[] neighborWeightCnt;
    delete[] m_neighbor;
    for (map <int, Cluster* >::iterator ci=result_clst.begin(); ci!=result_clst.end(); ci++)
      delete ci->second;
}

string FastGraphCluster::intToString(int a)
{
  stringstream ss;
  ss << a ;
  return ss.str();
}

bool FastGraphCluster::checkClst(Cluster * cl)
{
  int ai = cl->nodes.size(); // old version, -1
  if(ai < m_nLowerSize || ai*(ai-1)/2.0*m_dLowDen > (float) cl->n_edge)
    return false;
  return true;
}

void FastGraphCluster::copyClst(int old_id, int new_id)
{
  Cluster * i_clst = result_clst[old_id];
  result_clst[new_id] = i_clst;
  result_clst.erase(old_id);
  for (set<int>::iterator i=i_clst->nodes.begin(); i!=i_clst->nodes.end(); i++)
    clstID[*i] = new_id; 
  
}


void FastGraphCluster::deleteClst(ofstream& fout, int cid, Cluster * i_clst /*=NULL*/, bool printC /*= true*/, int oldEdge/*=0*/)
{
  if (i_clst==NULL) i_clst = result_clst[cid];
  if (printC==true) printClst(fout, cid, i_clst, oldEdge);  // first print out, then delete
  set<int>::iterator i;
  for (i=i_clst->nodes.begin(); i!=i_clst->nodes.end(); i++)
    clstID[*i] = -1; // dissolve from clusters
  delete i_clst;
  result_clst.erase(cid);
}

////  set< pair<int, int>, ComparePair > sortCount;
struct ComparePair
{
  bool operator() (const pair<int, int> a, const pair<int, int> b) const 
  {return a.second >= b.second;}
  ///{return a.first >= b.first;}
};

void mapToVec(map<int, int> &nodeCount, set< pair<int, int>, ComparePair > &sortCount)
{
  for (map <int, int>::iterator ni=nodeCount.begin();ni!=nodeCount.end();ni++)
    sortCount.insert(make_pair(ni->second, ni->first)); 
}

template <class K, class V> 
void addKey(map <K,V> &m, K key)
{
  typename map <K,V>::iterator it;
  if ((it=m.find(key)) == m.end()) {
    m[key] = 1;
  } else it->second++;
}


void FastGraphCluster::dissolve(ofstream& fout, vector< pair <int, int > > &delEdge)
{
  map<int, int > changed_clst;   // use struct {int count, set <int> nodes} to speed up 
  set<int> changed_clst_set;
  map<int, int > changed_nodes;
  Cluster* tmp_clst;
  int ci, ai, flag=1;
    
  for (vector< pair <int, int > >::iterator i=delEdge.begin();i!=delEdge.end();i++)
    {
      ci=clstID[(*i).first];
      if ( ci > -1 &&  ci==clstID[(*i).second]) // edges to delete within clusters
        {
	  flag = 0;
	  addKey(changed_clst, ci);
	  addKey(changed_nodes, (*i).first);
	  addKey(changed_nodes, (*i).second);
	}
    }  

  if (flag) return;
  for (map<int, int >::iterator cc = changed_clst.begin(); cc != changed_clst.end(); cc++)
    {
      tmp_clst = result_clst[cc->first];
      tmp_clst->n_edge -= cc->second;

      if (!checkClst(tmp_clst)) {
	deleteClst(fout, cc->first, tmp_clst, true, tmp_clst->n_edge + cc->second + tmp_clst->pre_del_num);
      }  else {
	tmp_clst->pre_del_num = cc->second;
	changed_clst_set.insert(cc->first); 
      }
    }       
  // remove nodes
  set <int> clst_map; // clusters be copied(to new cluster) and deleted, due to edgeChange & smallS
  set <int> clst_del; // clusters be deleted, due to smallS
  set <int>::iterator cm;
  map< int, EdgeInfo*>::iterator imap;
  size_t win0, win1;
  for (map<int, int>::iterator si=changed_nodes.begin(); si!=changed_nodes.end();si++)
    {
      ci = clstID[si->first];
      if (ci == -1) continue;
      tmp_clst = result_clst[ci];
      ai = 0;   // num of neighbors in the same cluster (in current window), m_neighbor is updated
      for (imap = m_neighbor[si->first].begin(); imap!=m_neighbor[si->first].end(); imap++)
	  if (clstID[imap->first]==ci) ai++;
      if (ai >= 3.0 && ai >= (tmp_clst->nodes.size()-1) * m_dLowerIncrease) continue;
      
      clstID[si->first] = -1; // remove this node from cluster      
      cm = clst_map.find(ci);
      if (cm == clst_map.end()) {
	printClst(fout, ci, tmp_clst);
	clst_map.insert(ci);
      }
      if (tmp_clst->nodes.size()-1 < m_nLowerSize){ // size too small, delete
	clst_del.insert(ci);
	deleteClst(fout, ci, tmp_clst, false);
      } else {                //remove this node,  copy rest to new cluster	    
	
        // wrong, reported num of edges is lower than truth. april 17. However, good approximation
	// to make it right, write down the neighbor id, calculate the exact number 
	tmp_clst->n_edge -= ai;  
	//tmp_clst->n_edge -= (ai-(si->second)); very wrong in some cases, n_edge shouldn't increase
	tmp_clst->nodes.erase(si->first); 	
      }           
    }
  // some cluster only has edge changed, copied to a new clst but itself not printed!
  clst_map.insert(changed_clst_set.begin(), changed_clst_set.end());
  for (cm = clst_map.begin(); cm!=clst_map.end(); cm++){
    if (clst_del.find((*cm))==clst_del.end()) {  // not deleted
      copyClst(*cm, clst_topindex);    
      tmp_clst = result_clst[clst_topindex];
      clstBorder_update(tmp_clst->nodes, win0, win1);  
      tmp_clst->pre_del_num = 0;
      tmp_clst->p_start = win0;
      tmp_clst->p_end = win1;
      clst_topindex++;
    }
  }
}

// add Node to existing methods
void FastGraphCluster::addNode(ofstream& fout, vector< pair <int, int > > &addEdge)
{
  int clst_a, clst_b;
  map <int, int> node_toappend;
  Cluster* tmp_clst;
  EdgeInfo* p_edge;
  map <pair <int, int>, int> clst_tojoin;

  if (addEdge.size()==0 || result_clst.size()==0) return;
  for (vector< pair <int, int > >::iterator i=addEdge.begin();i!=addEdge.end();i++)
    {
      clst_a = clstID[(*i).first];
      clst_b = clstID[(*i).second]; 
      if (clst_a == clst_b)
 	{
	  //cout << "new "  <<  clst_a << " " << clst_b << endl;
 	  if (clst_a == -1)
	    {
	      continue;
	    }
	  tmp_clst = result_clst[clst_a];
 	  tmp_clst->n_edge++;
	  p_edge = m_neighbor[(*i).first][(*i).second];
	  tmp_clst->p_start = getmax(tmp_clst->p_start, p_edge->p0);
	  tmp_clst->p_end = getmin(tmp_clst->p_end, p_edge->p1);
	  //cout << "samecluster " << endl; 
 	} else if (clst_a == -1 && clst_b > -1)
 	{
	  addKey(node_toappend, (*i).first);  
	  //cout << "clst" << endl;     
 	} else if (clst_b == -1 && clst_a > -1)
	{
	  addKey(node_toappend, (*i).second);
	  //cout << "clst" << endl;     
 	} else 
	{
	  if (clst_a > clst_b) addKey(clst_tojoin, make_pair(clst_a, clst_b));
	  else addKey(clst_tojoin, make_pair(clst_b, clst_a));
	  //cout << "missingEdge " << clst_a << " " << clst_b <<  endl;             
	}
    }
  
  map <pair <int, int>, int>::iterator ct;
  int sa, sb;
  set <int> clst_todel;
  for (ct=clst_tojoin.begin(); ct!= clst_tojoin.end(); ct++)
    {
      //if (ct->second >= (1-m_dLowDen) * result_clst[ct->first.first]->n_edge)
      if (ct->second >= CLST_TOJOIN * result_clst[ct->first.first]->n_edge)
	clst_todel.insert(ct->first.first);
      //if (ct->second >= (1-m_dLowDen) * result_clst[ct->first.second]->n_edge)
      if (ct->second >= CLST_TOJOIN * result_clst[ct->first.second]->n_edge)
	clst_todel.insert(ct->first.second);
    }
  
  //cout << "delete these clst " << clst_todel.size() << endl;
  for (set <int>::iterator i = clst_todel.begin(); i!= clst_todel.end(); i++)
    {
      tmp_clst = result_clst[*i];
      deleteClst(fout, *i, tmp_clst, true, tmp_clst->n_edge);
    }
  
  map<int, int>::iterator ai, cm;
  map<int,EdgeInfo*>::iterator imap;
  map <int, int> clst_count;
  set< pair <int, int>, ComparePair > clst_count_set;
  map <int, int> clst_map;
  int ci;
  map <int, size_t> clst0, clst1; // if node is added, min_spanning positions

  for (ai=node_toappend.begin(); ai!=node_toappend.end();ai++)  // node_toappend[node_id] = counts
    {
      if ( ai->second < 3.0) continue; // not necessary though
      for (imap = m_neighbor[ai->first].begin(); imap!=m_neighbor[ai->first].end(); imap++)
	{
	  if ((ci = clstID[imap->first]) > -1) {
	    addKey(clst_count, ci);   // a new clst_count for each node
	    update0(clst0, ci, (imap->second)->p0); 
	    update1(clst1, ci, (imap->second)->p1);
	  }
	}
      /* add sorting of clst_count, April 17 */      
      for (map<int, int>::iterator icl=clst_count.begin(); icl!=clst_count.end(); icl++)
	clst_count_set.insert(make_pair(icl->first, icl->second));
      
      set< pair <int, int>, ComparePair >::iterator bi;
      for (bi=clst_count_set.begin(); bi!=clst_count_set.end();bi++) 
	{
	  if ((*bi).second < 3.0) continue; // node ai's support to cluster < 3
	  tmp_clst = result_clst[(*bi).first];
	  if ((*bi).second >= tmp_clst->nodes.size()*m_dLowDen) 
	    {	      
	      if ((cm = clst_map.find((*bi).first))== clst_map.end()) // print old, add new cluster
		{
		  printClst(fout, (*bi).first, tmp_clst);
		  clst_map[(*bi).first] = clst_topindex++;
		}	      
	      /*
	      if (tmp_clst->p_end < clst0[(*bi).first])
		{ // dissolve didn't update position right
		  cerr << "check dissolve update positions " << endl;
		  exit(0);
		}	      
	      */
	      tmp_clst->nodes.insert(ai->first);
	      tmp_clst->n_edge += (*bi).second;
	      tmp_clst->p_start = getmax(tmp_clst->p_start, clst0[(*bi).first]);
	      tmp_clst->p_end   = getmin(tmp_clst->p_end,   clst1[(*bi).first]);
	      clstID[ai->first] = (*bi).first;  // use the old clustid
	      break;
	    }
	}
      clst_count.clear();
      clst_count_set.clear();
      clst0.clear();
      clst1.clear();
    }
  for (ai = clst_map.begin(); ai != clst_map.end(); ai++)
    copyClst(ai->first, ai->second);
}


void FastGraphCluster::printClst(ofstream& fout, int cn, Cluster * cl /* = NULL*/, int oldEdge/*=0*/)
{
  if (cl==NULL) cl = result_clst[cn];
  fout <<"clst" << cn << "\t" << cl->p_start << "\t" << cl->p_end << "\t"; 

  // for debug purpose, print out node number and edge number 
  /* 
  if (oldEdge > 0)
    fout <<"clst" << clstmode << cn << "\t" << cl->p_start << "\t" << cl->p_end << "\t" << oldEdge << "\t" << cl->nodes.size() << "\t"; 
  else
    fout <<"clst" << clstmode << cn << "\t" << cl->p_start << "\t" << cl->p_end << "\t" << cl->n_edge+cl->pre_del_num << "\t"<< cl->nodes.size() << "\t"; 
  */
  
  // for debug purpose, NB: cur_pos_start is not really the start pos, only end pos is correctly tracked
  /*  
  fout <<"clst" << clstmode << cn << "\t" << cl->p_start << "\t" << cl->p_end << "\t" << cur_pos_start << "\t"<< cur_pos << "\t"; 
  */

  for (set <int>::iterator j=cl->nodes.begin(); j!=cl->nodes.end(); j++)
    fout << vertexName[*j] << "\t";
  fout << endl;
}


void FastGraphCluster::printAllClst(ofstream& fout)
{
  for (map <int, Cluster* >::iterator ci=result_clst.begin(); ci!=result_clst.end(); ci++)
    {   
      Cluster * cl = ci->second;
      fout << "clst" << ci->first << "\t" << cl->p_start << "\t" << cl->p_end <<"\t" << cl->n_edge << "\t" << cl->nodes.size() << "\t";
      for (set <int>::iterator j=cl->nodes.begin(); j!=cl->nodes.end(); j++)
	fout << vertexName[*j] << "\t";
      fout << endl;
    }
}

void FastGraphCluster::updateInput(list<Pairmatch * > &active_matches)
{

  int i;
  float *nw = neighborWeightCnt;
  FiboNode *mp2;
  for (i=0;i < m_nVertex;i++)
    {
      mp2=m_pHeapNode2[i];         // use mp2 to reserve the space, alloc/dealloc expensive
      mp2->key.wsum = UNEXPLORED;
      mp2->key.esum = 0;
      if (clstID[i] > -1){  // already clustered
	m_pHeapNode[i] = NULL;
      } else m_pHeapNode[i] = mp2;
      *nw++ = 0;
    }
  list< Pairmatch * >::iterator am;  
  for (am = active_matches.begin(); am != active_matches.end(); am++)
    {
      if (clstID[(*am)->i1] > -1 || clstID[(*am)->i2] > -1) continue; 
      neighborWeightCnt[(*am)->i1] += (*am)->weight;
      neighborWeightCnt[(*am)->i2] += (*am)->weight;
    }
  float maxWeightDegree = 0;
  for (i=0;i<m_nVertex;i++){
    if( neighborWeightCnt[i] > maxWeightDegree ) maxWeightDegree = neighborWeightCnt[i];
  }
  //cerr << "weightCnt " << neighborWeightCnt[1835] << endl;
  seedArray = new DegreeArray(neighborWeightCnt,m_nVertex,maxWeightDegree);
}

void FastGraphCluster::fastCluster(ofstream& fout)
{
  int i=0;
  pair< size_t, size_t > plen;
  //cerr <<endl << "win1 " << neighborWeightCnt[1835] << endl;
  //////while (!seedArray->empty() && seedArray->top >= m_nLowerSize) ## notice this!!!
  while (!seedArray->empty() && seedArray->top >= m_nLowerSize-1)
    {
      i = seedArray->getMax();
      //cerr << "i" << i << ";" << neighborWeightCnt[i] << ";" << seedArray->top << ";";
      expand(i);
    }
  //cerr <<endl << "win2" << endl;
  if (!seedArray->empty())
    delete seedArray;
}

float FastGraphCluster::getWeight(float edgeWeight,float vertexWeight, float seedW)
{
  int nbin=10;      // 0.8, 0.9, 1.0 have different values
  return m_nVertex * edgeWeight + vertexWeight; // similar to dash
  //return  m_nVertex * (int)(nbin * edgeWeight) + vertexWeight; 
}

void FastGraphCluster::printNeighbor(int index)
{
  map<int, EdgeInfo*>::iterator imap;
  for (imap = m_neighbor[index].begin(); imap!=m_neighbor[index].end(); imap++)
    if ((imap->second)->weight > 0) cout << imap->first << " ";
  cout << endl;
}

int FastGraphCluster::expand(int index)
{
        int tEdge=0, imin, i,j,w3, nVertex = 0;
	set<int> result;
	float density, w_density,increase,w,totalWeight = 0,oldWeight;
	float *weightlist;

	FibonacciHeap heap;	// local expanding heap
	set <int> changed;

	FiboNode *ptr = m_pHeapNode[index];
	ptr->key.wsum = 0;
	ptr->key.esum = 0;
	heap.insert(ptr);

	// start second level heuristic value selection
	int maxindex = -1; // default 0 may cause problem
	float maxWeight = 0;

	map<int, EdgeInfo*>::iterator imap;    

    float seed_WeightCnt = neighborWeightCnt[index];
    //cerr << "start " << neighborWeightCnt[index] << " " << m_neighbor[index].size();
   for (imap = m_neighbor[index].begin(); imap!=m_neighbor[index].end(); imap++)
	{
        j = imap->first;
        if (m_pHeapNode[j] == NULL) continue;
	w = getWeight((imap->second)->weight,neighborWeightCnt[j], seed_WeightCnt);

        if (w > maxWeight )
	    {
            maxindex = j;
            maxWeight = (float)w;
	    }
    }
   
   if (maxindex < 0) {
     seedArray->remove(index,neighborWeightCnt[index]);
     return 0;
   }
   //cerr << "start1;"<<index <<";"<<maxindex<<";";

    // for the expanding procedure, definitely select this edge first
    m_neighbor[index][maxindex]->weight += maxWeight;

    while (!heap.empty())
	  {
	    ptr = heap.extractMin();
	    imin = ptr->key.index;
	    w = ptr->key.wsum;   // initial value = 0
	    w3 = ptr->key.esum;

	    totalWeight += w;  // initial value = 0
	    tEdge += w3;
	    nVertex ++;

	    // more than first 2 seeds, check density etc. 
	    if (nVertex > 1)
	      {
		w_density = 2.0*totalWeight/(nVertex*(nVertex-1));
		density = 2.0*tEdge/(nVertex*(nVertex-1));
		increase = w3/(density*(nVertex-1));  // increase of edge

		///if ( density < m_dLowDen || increase < m_dLowerIncrease || w_density < MIN_CLST_WEIGHT)
		if ( density < m_dLowDen || increase < m_dLowerIncrease )
		  {
		    // about 1/3 of them stopped because increase < 0.5
		    //cout << "break " << density << " "<< increase << " "<< w_density << " " << nVertex << endl;
		    nVertex--;
		    totalWeight -= w;
		    tEdge -= w3;
		    break;	// exceeding the lowest density
		  }
	      }

	    m_pHeapNode[imin] = NULL;
	    // remove merged vertex from seed list
	    seedArray->remove(imin,neighborWeightCnt[imin]);
	    result.insert(imin);	// write back result

    for (imap = m_neighbor[imin].begin(); imap!=m_neighbor[imin].end(); imap++)
    {
        j = imap->first;
        w = (imap->second)->weight;
        ptr = m_pHeapNode[j];
        if (ptr==NULL) continue;
	if (ptr->key.wsum == UNEXPLORED)
	  {	// a new expanded vertex
                ptr->key.wsum = w;
                ptr->key.esum = 1;
                heap.insert(ptr);
            }else  heap.decrease(ptr,HeapNode( ptr->key.wsum+w,j,ptr->key.esum+1));
            // decrease its weight degree which is independent of the cluster
		    oldWeight = neighborWeightCnt[j];
            
		    if ( ( nVertex == 1 ) && (j == maxindex) )
            {
                seedArray->decrease(j,oldWeight,oldWeight + maxWeight - w);
                neighborWeightCnt[j] -= (w-maxWeight);
            }else
            {
                seedArray->decrease(j,oldWeight,oldWeight - w);
                neighborWeightCnt[j] -= w;
            }
		    changed.insert(j);
    }

          
	if (nVertex == 1 ) totalWeight -= maxWeight; // deduce back extra weight for seeding pair
    }
    for ( set<int>::iterator iter = changed.begin() ; iter != changed.end() ; iter++ )
      {
	i = *iter;
	if ( m_pHeapNode[i] != NULL ){
	  m_pHeapNode[i]->key.wsum = UNEXPLORED;
	  m_pHeapNode[i]->key.esum = 0;
	}
      }
    changed.clear();
    m_neighbor[index][maxindex]->weight -= maxWeight;

    if (result.size() >= m_nLowerSize)
      {
	size_t win0, win1;
    	for (set<int>::iterator ni=result.begin();ni!=result.end();ni++)
	  clstID[*ni] = clst_topindex;
	clstBorder(result, win0, win1);
	// cout << "new cluster " << clst_topindex <<" " << tEdge << endl;
	result_clst[clst_topindex++] = new Cluster(result, tEdge, win0, win1);
      }
    return 1;
}

void FastGraphCluster::clstBorder(set<int> &clst, size_t &pa, size_t &pb)
{
  set <int>::iterator it,jt;
  map <int, EdgeInfo*>::iterator mp;
  size_t ia=0, ib=0, ta, tb;
  int sa = 0, sb = clst.size();
  EdgeInfo *p_edge;
  for (it=clst.begin(); sa < sb-2; it++)
    {
      jt = it;
      jt++;
      sa++;
      for ( ; jt!=clst.end(); jt++)
	{ 
	  /// if ((p_edge = m_neighbor[*it][*jt])==NULL) continue; this is very wrong!
	  if ( (mp = m_neighbor[*it].find(*jt)) ==  m_neighbor[*it].end()) {
	    continue;
	  } else p_edge = mp->second;
	  if (ia==0) {
	    ia = p_edge->p0;
	    ib = p_edge->p1;
	    ///ib = getmax(p_edge->p1, cur_pos - WINDOW_SIZE); // should bigger than window start
	  } else {
	    ia = getmax(ia, p_edge->p0);
	    ib = getmin(ib, p_edge->p1);
	  }
	}
    }
  pa = ia;
  pb = ib;
}

void FastGraphCluster::clstBorder_update(set<int> &clst, size_t &pa, size_t &pb)
{
  set <int>::iterator it,jt;
  map <int, EdgeInfo*>::iterator mp;
  size_t ia=0, ib=0, ta, tb;
  int sa = 0, sb = clst.size();
  EdgeInfo *p_edge;
  for (it=clst.begin(); sa < sb-2; it++)
    {
      jt = it;
      jt++;
      sa++;
      for ( ; jt!=clst.end(); jt++)
	{ 
	  /// if ((p_edge = m_neighbor[*it][*jt])==NULL) continue; this is very wrong!
	  if ( (mp = m_neighbor[*it].find(*jt)) ==  m_neighbor[*it].end()) continue;
	  else p_edge = mp->second;
	  if (p_edge->p1 < cur_pos) continue; 	    
	  if (ia==0) {
	    ia = p_edge->p0;
	    ib = p_edge->p1;
	  } else {
	    ia = getmax(ia, p_edge->p0);
	    ib = getmin(ib, p_edge->p1);
	  }
	}
    }
  pa = ia;
  pb = ib;
}

bool HeapNode::operator >(const HeapNode &right)
{
	return wsum<right.wsum;
}

bool HeapNode::operator <(const HeapNode &right)
{
	return wsum>right.wsum;
}

bool HeapNode::operator <=(const HeapNode &right)
{
	return wsum >= right.wsum;
}

bool HeapNode::operator >=(const HeapNode &right)
{
	return wsum <= right.wsum;
}
