#pragma once
#include "emi.h"
#include <utility>
#include <string>
#include <fstream>
#include <iostream>
#include <vector>
#include <set>
#include <sstream>
#include <list>
#include <map>

using namespace std;

class FiboNode;
class DegreeArray;
class Pairmatch; //forward declaration
class Cluster;
class EdgeInfo;

// node used in heuristic vertex adding
class HeapNode{
public:
	float wsum;	// how many edges weight are connected to current module
	int index;	// index of the vertex
	int esum;
        HeapNode( float wsum = 0 , int index = 0 , int esum = 0):wsum(wsum),index(index),esum(esum){}

	// the order here are totally reversed. we will get a max heap instead
	bool operator >(const HeapNode &right);
	bool operator >=(const HeapNode &right);
	bool operator <(const HeapNode &right);
	bool operator <=(const HeapNode &right);
};

class FastGraphCluster
{
 public:
  FastGraphCluster(float density, int lowersize,float lowerincrease, int m_nVertex);
  ~FastGraphCluster(void);
  void updateInput(list<Pairmatch * > &active_matches);
  void fastCluster(ofstream& fout);
  //map <int, float> *m_neighbor;
  map <int, EdgeInfo* > *m_neighbor;
  map <int, Cluster* > result_clst;

  int clst_topindex;
  void printAllClst(ofstream& fout);
  void dissolve(ofstream& fout, vector< pair <int, int > > &delEdge);
  void addNode(ofstream& fout, vector< pair <int, int > > &addEdge);

private:
	// starting from a local seed posize_t and heuristically get the local dense module
        void printNeighbor(int index);
        int expand(int index);
	float getWeight(float edgeweight,float vertexweight, float seedW);

	FiboNode **m_pHeapNode;	// we will use Fibonacci heap
	FiboNode **m_pHeapNode2;
	int m_nVertex,m_nLowerSize;	// lower size of the modules;
	DegreeArray *seedArray;	// extract min + decrease key
	// parameters for the algorithm searching
	float m_dLowDen,m_dLowerIncrease;	// lower density of the modules
	float *neighborWeightCnt;	// total sum of the weight for  neighbor
	string intToString(int a);	
	map<int, int > clstID;
	
	void clstBorder(set<int> &clst, size_t &pa, size_t &pb);
	void clstBorder_update(set<int> &clst, size_t &pa, size_t &pb);
	void deleteClst(ofstream& fout,int cid, Cluster * i_clst=NULL, bool printC=true, int oldEdge=0);
	void copyClst(int old_id, int new_id);
	void printClst(ofstream& fout, int cn, Cluster * cl = NULL, int oldEdge=0);
	bool checkClst(Cluster * cl); // check if density and size satisfy
};
